""" @namespace python_rpnifs.expr
    @brief Handles expression generation & mutation
"""
import copy
import random

import numpy

import pyrpn

class RpnSym(object):
    """ @brief Abstract class representing RpnExpr symbols """
    
    def __init__(self):
        raise NotImplementedError('Abstract class')

    def __sub__(self, b):
        """ @return the difference between two symbols : 2 if types differs,
            1 if same type but different value
        """
        if type(b) != type(self):
            return 2
        elif str(b) !=  str(self):
            return 1
        return 0

class RpnOp(RpnSym):
    """ @brief an RPN operation symbol """
    
    def __init__(self, op=None):
        """ @brief instanciate an RPN symbol
            @param op if None a random op is choosen else a value from
                pyrpn.get_ops().values() is expected
        """
        if op is None:
            op = self.random()
        elif op not in pyrpn.get_ops().values():
           raise ValueError('Invalid operation %r' % op)
        self.__op = op

    def __str__(self):
        return self.__op

    def __copy__(self):
        return type(self)(self.__op)

    def to_bytes(self):
        """ @return a 9 bytes representation of the symbol """
        brepr = self.__op.encode('utf8')
        return bytes([1] + ([0]*(8-len(brepr)))) + brepr

    @staticmethod
    def random():
        """ @return a random RpnOp instance """
        return random.choice(list(pyrpn.get_ops().values()))


class RpnConst(RpnSym):
    """ @brief an RPN numeric constant symbol """
    
    def __init__(self, value=None):
        """ @brief instanciate an RPN numeric constant symbol
            @param value a 64bit unsigned integer (a random one is choosen
                if None given)
        """
        if value is None:
            value = random.randint(0, (1<<64)-1)
        elif value < 0 or value > (1<<64)-1:
            raise ValueError('Outof bound value %d' % value)
        self.__val = value

    def __copy__(self):
        return type(self)(self.__val)

    def __str__(self):
        return '0x%X' % self.__val

    def to_bytes(self):
        """ @return a 9 bytes representation of the symbol """
        return bytes([0]) + self.__val.to_bytes(8, 'little')


class RpnVar(RpnSym):
    """ @brief an RPN variable symbol
        
        Variables are indexed with integer, string representation are : A0, A1
        etc.
    """

    def __init__(self, varnum=None, nvar=1):
        """ @brief instanciate an RPN variable symbol
            @param varnum the variable number (random one choose if None given)
            @param nvar the maximum variable number
        """
        if varnum is None:
            if nvar == 1:
                varnum = 0
            else:
                varnum =  random.randint(0, nvar-1)
        elif varnum < 0 or varnum >= nvar:
            raise ValueError('Variable A%d do not exists' % varnum)
        self.__val =  varnum
        self.__nvar = nvar
    
    @property
    def nvar(self):
        """ @return the maximum variable number """
        return self.__nvar

    def __copy__(self):
        return type(self)(self.__val, self.__nvar)

    def __str__(self):
        return 'A%d' % self.__val

    def to_bytes(self):
        """ @return a 9 bytes representation of the symbol """
        brepr = str(self).encode('utf8')
        return bytes([1] + ([0]*(8-len(brepr)))) + brepr


class RpnExpr(object):
    """ @brief Wrapper on pyrpnifs.RPNExpr
        
        Allow using pyrpnifs.RPNExpr while keeping a high level representation
        of the expression (using @ref RpnSym child classes).
    """

    def __init__(self, sz=0, w_op=4, w_const=1, w_var=2, nvar=1):
        """ @param sz The expression size for random generation
            @param w_op RpnOp weight on random generation
            @param w_const RpnConst weight on random generation
            @param w_var RpnVar weight on random generation
            @param n_var the variable count
        """
        self.sz = 0
        self._weight = (w_op, w_const, w_var)
        choices = (RpnOp, RpnConst, lambda: RpnVar(nvar=nvar))
        self.expr = [random.choices(choices, self._weight)[0]()
                       for _ in range(sz)]
        self.__nvar = nvar
        if sz > 0:
            self.refresh_expr()

    def refresh_expr(self):
        """ @brief Refresh the internal pyrpn.RPNExpr using the
                @ref RpnSym representation
        """
        self.pyrpn = pyrpn.RPNExpr(str(self), self.__nvar)

    def __str__(self):
        return ' '.join([str(elt) for elt in self.expr])

    def __repr__(self):
        return '<RpnExpr sz=%d %r>' % (len(self.expr), str(self))

    def __copy__(self):
        w_op, w_const, w_var = self._weight
        ret = RpnExpr(0, w_op, w_const, w_var, self.__nvar)
        ret.expr = [copy.copy(elt) for elt in self.expr]
        ret.refresh_expr()
        return ret

    def to_bytes(self):
        """ @return a bytes representation with 9 bytes for each symbols """
        return b''.join([elt.to_bytes() for elt in self.expr])

    def __sub__(self, b):
        return self.levenshtein(b)

    def eval(self, *args):
        """ @brief Evaluate the expression with given arguments
            @param args* an array of argument (with len  = nvar)
        """
        return self.pyrpn.eval(*args)

    def mutation(self, n=1, min_len=3, w_add=1.25, w_del=1, w_mut=2,
                 w_mut_soft=4, w_add_elt=(1,1,1), w_mut_elt=(1,1,1)):
        """ @brief Expression mutation
            @param n mutation count
            @param min_len the minimum length for an expression
            @param w_add weight of adding a symbol
            @param w_del weight of deleting a symbol
            @param w_mut weight of mutating a symbol
            @param w_mut_soft weight of mutationg a symbol without type change
            @param w_add_elt weights for each types (op, const, var) when
                adding a symbol
            @param w_mut_elt weight for each types (op, const, var) when
                mutating a symbol
            @note The internal pyrpn.RPNExpr is refreshed after mutations
        """
        for _ in range(n):
            if len(self.expr) <= min_len:
                self.add(*w_add_elt)
                return
            funs = ((self.add, w_add_elt),
                    (self.delete, []),
                    (self.mutate, w_mut_elt),
                    (self.mutate_soft, []))
            choice = random.choices((0,1,2),
                                   (w_add, w_del, w_mut))[0]
            fun, args = funs[choice]
            fun(*args)
        self.refresh_expr()
        pass

    def add(self, w_op=1, w_const=1, w_var=1):
        """ @brief Mutate by adding a symbol """
        idx = random.randint(0,len(self.expr))
        new = random.choices((RpnOp, RpnConst, RpnVar), self._weight)[0]
        new = new(nvar=self.__nvar) if new == RpnVar else new()
        self.expr.insert(idx, new)
        pass
    
    def delete(self, *args):
        """ @brief Mutate by deleting a symbol """
        idx = random.randint(0, len(self.expr)-1)
        self.expr.pop(idx)
        pass

    def mutate(self, w_op=1, w_const=1, w_var=1):
        """ @brief Mutate changing a symbol into another """
        idx =  random.randint(0, len(self.expr)-1)
        new = random.choices((RpnOp, RpnConst, RpnVar), self._weight)[0]
        new = new(nvar=self.__nvar) if new == RpnVar else new()
        self.expr[idx] = new
        pass
    
    def mutate_soft(self):
        """ @brief Mutate changing a symbol into another of the same type """
        idx =  random.randint(0, len(self.expr)-1)
        cur = self.expr[idx]
        cls = type(cur)
        if cls == RpnVar:
            new = cls(nvar=cur.nvar)
        else:
            new = cls()
        self.expr[idx] =  new
        pass

    def levenshtein(a, b, custom_weight=True):
        """ @brief Calculate a levenshtein distance between two expression
            @param a expression A
            @param b expression B
            @param custom_weight If True a different symbol of same type weight
                1 and a different symbol of different type weight 2. Else no
                special weight is applied.
            @return an integer
        """

        len_a = len(a.expr)
        len_b = len(b.expr)

        dist = numpy.zeros((len_a+1, len_b+1))
        for i in range(len_a):
            dist[i][0] = i
        for i in range(len_b):
            dist[0][i] = i

        for ia in range(len_a):
            for ib in range(len_b):
                diff = a.expr[ia] - b.expr[ib] # in (0,1,2)
                if diff > 0:
                    val = min([dist[ia][ib], dist[ia+1][ib], dist[ia][ib+1]])
                    dist[ia+1][ib+1] = val + (diff if custom_weight else 1)
                else:
                    dist[ia+1][ib+1] = dist[ia][ib]

        return dist[len_a][len_b]

    @classmethod
    def from_string(cls, expr, nvar=1):
        """ @brief Return an RpnExpr instance given a string representing
            an expression
            @param epxr the string representation
            @param nvar the variable count
        """
        res = RpnExpr(sz=0)

        for spl in expr.split(' '):
            spl = spl.strip()
            if not spl:
                continue
            if spl[0] == 'A':
                try:
                    argnum = int(spl[1:])
                except Exception:
                    raise ValueError('Invalid argument %r' % spl)
                self.expr.append(RpnVar(argnum, nvar))
            elif spl in pyrpn.get_ops().values():
                self.expr.append(RpnOp(spl))
            else:
                try:
                    val = int(spl)
                except Exception:
                    raise ValueError('Invalid token %r' % spl)
                self.expr.append(RpnVal(spl))
        return res
        
