#!/usr/bin/python3
# Copyright 2020 Weber Yann
#
# This file is part of geneifs.
#
#        geneifs is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        geneifs is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with geneifs.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import random
import math
import pickle

import unittest
from unittest import skip

IMAX = (1<<64)-1

try:
    import pyrpn
except (ImportError, NameError) as e:
    print("Error importing pyrpn. Try to run make.",
          file=sys.stderr)
    raise e

## TODO : move this function somewhere or implement it in C
def decode_state(state):
    """ Return a dict containing rpn state dump fields
    """
    res = dict()
    res['real_sz'] = len(state)
    res['total_sz'] = int.from_bytes(state[0:8], byteorder="little")
    res['argc'] =  int.from_bytes(state[8:16], byteorder="little")
    res['stack_sz_bytes'] = state[16:24]
    res['stack_sz'] = state[16]
    res['token_sz'] = int.from_bytes(state[24:32], byteorder="little", signed=True)
    res['stack'] = [int.from_bytes(state[i:i+8], byteorder="little")
                    for i in range(32, 32 + (8*res['stack_sz']),8)]
    res['tokens_off'] = 32 + (8*res['stack_sz'])
    res['tokens_real_sz'] = res['real_sz'] - res['tokens_off']
    res['tokens'] = [state[i:i+24]
                      for i in range(res['tokens_off'],
                                     res['tokens_off'] + (24*res['token_sz']),
                                     24)]
    return res
    pass


class Test0RpnModule(unittest.TestCase):

    def test_init(self):
        """ RPNExpr instanciation """
        expr = pyrpn.RPNExpr("42", 2)

    def test_init_badargs(self):
        """ RPNExpr instanciation with bad arguments """
        badargs = [('', 2), (), ('ab+',), ('ab+',300), (42, 42), ('ab', '42')]
        for badarg in badargs:
           with self.assertRaises((ValueError, TypeError)):
            expr = pyrpn.RPNExpr(*badarg)

    def test_init_loop(self):
        """ Testing pyrpn.RPNExpr multiple instanciation """
        exprs = []
        for i in range(256):
            expr = pyrpn.RPNExpr("42", 2)

    def test_init_args(self):
        """ Testing pyrpn.RPNExpr instanciation with arguments """
        for argc in range(0, 256):
            with self.subTest("RPNExpr('42', %d)" % argc):
                expr = pyrpn.RPNExpr("42", argc)

    def test_init_stack_sz(self):
        """ Instanciate RPNExpr specifiing a stack_size """
        for argc in range(0,256, 8):
            for sz in range(4, 256, 2):
                with self.subTest("RPNExpr('42', %d,%d)" % (argc, sz)):
                    expr = pyrpn.RPNExpr("42", argc, sz)

    def test_op_dict(self):
        """ Testing RPNExpr.get_ops() method """
        known_ops = dict([
            ('add', '+'),
            ('sub', '-'),
            ('mul', '*'),
            ('div', '/'),
            ('mod', '%'),
            ('neg', '!'),
            ('not', '~'),
            ('and', '&'),
            ('or', '|'),
            ('xor', '^'),
            ('>>', 'r'),
            ('<<', 'l'),
            ('xchg', 'x'),
            ('dup', 'd'),
            ('pop', 'p'),
        ])
        for long, short in pyrpn.get_ops().items():
            self.assertIn(long, known_ops)
            self.assertEqual(short, known_ops[long])

    def test_rand_expr(self):
        """ Testing RPNExpr.random_expr() """
        result = {}
        counters = {'vals': 0, 'args': 0}
        all_count = 0
        for sz in range(1,64,3):
            for argc in range(1,128):
                rnd_expr = pyrpn.random_expr(argc, sz)
                #print(repr(rnd_expr))
                for tok in rnd_expr.split(' '):
                    all_count += 1
                    if len(tok) == 0:
                        continue
                    try:
                        itok = int(tok, 0)
                        counters['vals'] += 1
                        continue
                    except Exception:
                        pass
                    if tok[0] == 'A':
                        counters['args'] += 1
                    else:
                        if tok not in counters:
                            counters[tok] = 0
                        counters[tok] += 1
        all_ops = len(pyrpn.get_ops()) + 2
        entropy = 1-sum([(n/all_count)**2
                          for _, n in counters.items()])
        self.assertGreater(entropy, 1-(1/all_ops), "Low entropy !")
    
    def test_pickle_state(self):
        """ Testing pickling/unpickling """
        e = pyrpn.RPNExpr('0x42 + A0', 2)
        e2 = pickle.loads(pickle.dumps(e))
        ds_e = decode_state(e.__getstate__())
        ds_e2 = decode_state(e2.__getstate__())
        self.assertEqual(e.__getstate__(), e2.__getstate__(),
                         msg="EXPR: %r != %r (%r != %r)" % (e, e2, ds_e, ds_e2))
        for i in range(100):
            e = pyrpn.RPNExpr(pyrpn.random_expr(0), 2)

            e2 = pickle.loads(pickle.dumps(e))

            self.assertEqual(e.__getstate__(), e2.__getstate__(),
                             msg="EXPR#%d : %r != %r" % (i,e, e2))

            e3 = pickle.loads(pickle.dumps(e2))

            self.assertEqual(e.__getstate__(), e3.__getstate__(), msg=e)
            self.assertEqual(e2.__getstate__(), e3.__getstate__(), msg=e)

if __name__ == '__main__':
    unittest.main()
